from datetime import datetime, timedelta
import logging
import os

GREEN = '\033[92m'
RED = '\x1b[31;21m'
RESET_SEQ = "\033[0m"


def print_success(msg, end=''):
    """Print a message to the console with green colored text"""
    if end != '' and end != '\n':
        print("                      ", end=end)
    print(f'{GREEN}{msg}{RESET_SEQ}', end=end)


def print_failure(msg, end=''):
    """Print a message to the console with red colored text"""
    if end != '':
        print("                      ", end=end)
    print(f'{RED}{msg}{RESET_SEQ}', end=end)


today = datetime.now()
yesterday = datetime.now() - timedelta(days=1)

# https://airflow.apache.org/docs/apache-airflow/stable/macros-ref.html
def datetime_to_ds(dt):
    """Convert a Python datetime object to a string with the format YYYY-MM-DD"""
    return dt.strftime('%Y-%m-%d')


def datetime_to_ds_nodash(dt):
    """Convert a Python datetime object to a string with the format YYYYMMDD"""
    return dt.strftime('%Y%m%d')


def get_default_params():
    """Return dict object containing keys/values that mimic Apache Airflow default macros"""
    return {
        'ds': datetime_to_ds(yesterday),
        'ds_nodash': datetime_to_ds_nodash(yesterday),
        'tomorrow_ds_nodash': datetime_to_ds_nodash(today),
        'tomorrow_ds': datetime_to_ds(today)
    }


def read_sql_file(file_path):
    try:
        # todo check if file ends with .sql
        if os.path.isfile(file_path):
            f = open(file_path, "r")
            config_file_content = f.readlines()

            no_comment_sql = ''
            for line in config_file_content:
                comment_index = line.find('--')
                if comment_index != -1:
                    no_comment_sql += line[:comment_index]
                else:
                    no_comment_sql += line
            return no_comment_sql
        else:
            raise ValueError(f'Error: File does not exist: {file_path}')
    except Exception as e:
        logging.error(e)
        return False


# @retry.with_exponential_backoff(
#       num_retries=MAX_RETRIES,
#       retry_filter=retry_on_http_timeout_and_value_error)
#   def _query_with_retry(self):
#     """Run Bigquery query with retry if got error http response"""
#     _LOGGER.info('Attempting to perform query %s to BQ', self.query)
#     # Create client here since it throws an exception if pickled.
#     bigquery_client = bigquery.Client(self.project)
#     query_job = bigquery_client.query(self.query)
#     rows = query_job.result(timeout=60)
#     return [row.values() for row in rows]


import json
import logging
import os
from pathlib import Path
import warnings

from google.cloud import bigquery
from jinja2 import Template

from bigquery_validator.bigquery_validator_util import get_default_params, print_success, print_failure, \
    read_sql_file, RESET_SEQ

warnings.filterwarnings("ignore", "Your application has authenticated using end user credentials")


class BigQueryValidator:
    """Convert Jinja templated sql query to a regularly formatted query and validate it using BigQuery dry run.
    Queries can be inputted as a string or by passing the file path to a sql file.
    """

    def __init__(self,
                 return_query_cost_as_dict=False,
                 use_query_cache=False):
        self.bq_client = bigquery.Client()
        self.params = self.load_params()
        self.use_query_cache = use_query_cache
        self.return_query_cost_as_dict = return_query_cost_as_dict


    def load_params(self):
        """Load default params"""
        params = get_default_params()

        # load params from global config
        def load_config_from_file(config_path):
            if os.path.isfile(config_path):
                try:
                    f = open(config_path, "r")
                    config_file_content = f.read()
                    config_params = json.loads(config_file_content)
                    logging.info(f'Loaded params from global config file: {config_path}')
                    return config_params
                except Exception as e:
                    logging.error('Error reading params from global config file. No global params will be loaded.')
                    return {}
            else:
                return {}

        home = str(Path.home())
        global_config_path = os.path.join(home, '.python_bigquery_validator/config.json')
        global_config_params = load_config_from_file(global_config_path)
        params = {**params, **global_config_params}

        # load params from local config
        local_config_path = os.path.join(os.getcwd(), 'bq_validator_config.json')
        local_config_params = load_config_from_file(local_config_path)
        params = {**params, **local_config_params}
        return params

    def render_templated_query(self, templated_query):
        """Convert the Jinja templated SQL to a valid query"""
        templated_query = templated_query.replace('params.', '')  # need this to get formatting correct
        t = Template(templated_query)
        rendered_query = t.render(self.params)
        rendered_query = rendered_query.replace('\n', ' ')
        return rendered_query

    def render_templated_query_from_file(self, file_path):
        """Convert the Jinja templated SQL to a valid query"""
        templated_query = read_sql_file(file_path)
        return self.render_templated_query(templated_query)

    # # todo finish
    # def parameterize_sql(self, query):
    #     default_params = get_default_params()
    #     for k, v in self.params.items():
    #         # default params do not require 'param.' prefix
    #         if k in default_params:
    #             query = query.replace(f'{{{{  {k} }}}}', v)
    #         else:
    #             query = query.replace(f'{{{{ params.{k} }}}}', v)
    #     return query

    def dry_run_query(self, query):
        """Run a BigQuery query with dry_run set to True.
        If the query succeeds it is valid and will return the estimated processing bytes required for the query.
        An exception will be thrown if the query is not valid.
        """
        try:
            job_config = bigquery.QueryJobConfig(dry_run=True, use_query_cache=self.use_query_cache)

            # Start the query, passing in the extra configuration.
            query_job = self.bq_client.query(
                query,
                job_config=job_config,
            )  # Make an API request.

            # A dry run query completes immediately.
            total_bytes = query_job.total_bytes_processed
            logging.info("This query will process {} bytes.".format(total_bytes))

            byte = 1
            kilobyte = 1024
            megabyte = 1024 * 1024
            gigabyte = 1024 * 1024 * 1024
            terabyte = 1024 * 1024 * 1024 * 1024

            query_cost = {
                'b': round(total_bytes / byte, 2),
                'kb': round(total_bytes / kilobyte, 2),
                'mb': round(total_bytes / megabyte, 2),
                'gb': round(total_bytes / gigabyte, 2),
                'tb': round(total_bytes / terabyte, 2)
            }

            if total_bytes > terabyte:
                rounded_total = query_cost['b']
                byte_type = 'TB'
            elif total_bytes > gigabyte:
                rounded_total = query_cost['gb']
                byte_type = 'GB'
            elif total_bytes > megabyte:
                rounded_total = query_cost['mb']
                byte_type = 'MB'
            elif total_bytes > kilobyte:
                rounded_total = query_cost['kb']
                byte_type = 'KB'
            else:
                rounded_total = query_cost['b']
                byte_type = 'B'

            if self.return_query_cost_as_dict:
                return True, query_cost
            else:
                message = f'This query will process {rounded_total} {byte_type}.'
                return True, message

        except Exception as e:
            error_string = str(e)
            error_minus_job_url = error_string.split('jobs?prettyPrint=false:')
            split_error = error_minus_job_url[1].split('\n\n')
            syntax_error = split_error[0]
            return False, syntax_error


    # TODO: validate output of query
    # Stolen from Apache Beam
    # https://github.com/apache/beam/blob/87e11644c44a4c677ec2faa78f50cdffbb33605a/sdks/python/apache_beam/io/gcp/tests/bigquery_matcher.py
    # @retry.with_exponential_backoff(
    #       num_retries=MAX_RETRIES,
    #       retry_filter=retry_on_http_timeout_and_value_error)
    #   def run_query(self):
    #     """Run Bigquery query with retry if got error http response"""
    #     _LOGGER.info('Attempting to perform query %s to BQ', self.query)
    #     # Create client here since it throws an exception if pickled.
    #     bigquery_client = bigquery.Client(self.project)
    #     query_job = bigquery_client.query(self.query)
    #     rows = query_job.result(timeout=60)
    #     return [row.values() for row in rows]

    def validate_query(self, templated_query):
        """Check if query passed as parameter is valid. If the query contains any Jinja templated params they will
        be converted to the associated param value if one exists.

        Parameters:
        templated_query (str): SQL query to be validated
        """
        try:
            formatted_query = self.render_templated_query(templated_query)
            querv_is_valid, message = self.dry_run_query(formatted_query)
            logging.info(f'Query is { "valid" if querv_is_valid else "invalid"}. {message}')
            return querv_is_valid, message
        except Exception as e:
            logging.error(e)
            return False, f"An error occurred while validating query - {templated_query}"

    def validate_query_from_file(self, file_path):
        """Same as validate_query() but reads query from a file rather than accepting it as a param

        Parameters:
        file_path (str): Path to the sql file on the file system
        """
        try:
            # todo check if file ends with .sql
            if os.path.isfile(file_path):
                templated_query = read_sql_file(file_path)
                return self.validate_query(templated_query)
            else:
                raise ValueError(f'Error: File does not exist: {file_path}')
        except Exception as e:
            logging.error(e)
            return False


    def auto_validate_query_from_file(self, file_path):
        """Continuously monitor a sql file and automatically validate the sql on every saved change to the file.
        Any Jinja templated params will be automatically parsed on update.

        Parameters:
        file_path (str): Path to the sql file on the file system
       """
        try:
            _cached_stamp = 0
            while True:
                # TODO monitor query validator config for changes too
                stamp = os.stat(file_path).st_mtime
                if stamp != _cached_stamp:
                    print(f'Loading...{RESET_SEQ}', end='                                                          \r')
                    _cached_stamp = stamp
                    templated_query = read_sql_file(file_path)
                    formatted_query = self.render_templated_query(templated_query)
                    querv_is_valid, message = self.dry_run_query(formatted_query)

                    logging.info(f'Query is {"valid" if querv_is_valid else "invalid"}')

                    if querv_is_valid:
                        # Extra white space here is quick workaround to remove all text from last message
                        # when the previous message length exceeds the new message length
                        print_success(f'Valid query. {message}{RESET_SEQ}',
                                      end='                                                                         \r')
                    else:
                        print_failure(f'Invalid query. {message}{RESET_SEQ}',
                                      end='                                                                         \r')
        except Exception as e:
            logging.error(e)
            return False



import unittest

from bigquery_validator.bigquery_result import BigQueryResult


class BigqueryResultTest(unittest.TestCase):

    def test_query_executes_by_default(self):
        query = "SELECT count(*) AS nrows FROM `{{ params.project }}.samples.github_timeline`"

        bqr = BigQueryResult(query)
        print(bqr.result)
        self.assertIsNotNone(bqr.result)

    def test_query_auto_executes_set_to_false_returns_empty_array(self):
        query = "SELECT count(*) AS nrows FROM `{{ params.project }}.samples.github_timeline`"

        bqr = BigQueryResult(query, auto_execute=False)
        self.assertEquals(bqr.result, [])

    def test_query_metadata_is_not_none(self):
        query = "SELECT count(*) AS nrows FROM `{{ params.project }}.samples.github_timeline`"

        bqr = BigQueryResult(query)
        result_metadata = bqr.metadata()
        unique_rows = len(result_metadata['unique_values']['nrows'])
        total_rows = result_metadata['nrows']
        self.assertIsNotNone(result_metadata)
        self.assertEquals(unique_rows, total_rows)

    def test_query_from_file_metadata_is_not_none(self):
        bqr = BigQueryResult(file_path='./sql/bigquery_result_metadata.sql')
        result_metadata = bqr.metadata()
        unique_rows = len(result_metadata['unique_values']['nrows'])
        total_rows = result_metadata['nrows']
        self.assertIsNotNone(result_metadata)
        self.assertEquals(unique_rows, total_rows)

    def test_query_metadata_returns_correct_columns(self):
        query = '''
        select 'andrew' as name, 21 as age
        union all
        select 'james' as name, 20 as age
        '''

        bqr = BigQueryResult(query)
        result_metadata = bqr.metadata()
        columns = result_metadata['columns']
        print(columns)
        self.assertEquals(columns, ['name', 'age'])

    def test_query_metadata_returns_correct_unique_values(self):
        query = '''
        select 'andrew' as name, 21 as age
        union all
        select 'james' as name, 20 as age
        '''

        bqr = BigQueryResult(query)
        result_metadata = bqr.metadata()
        unique_names = result_metadata['unique_values']['name']
        self.assertEquals(unique_names, ['andrew', 'james'])

    def test_query_from_file_metadata_returns_correct_unique_values(self):
        bqr = BigQueryResult(file_path='./sql/bigquery_result_test.sql')
        result_metadata = bqr.metadata()
        unique_names = result_metadata['unique_values']['name']
        self.assertEquals(unique_names, ['john', 'peter', 'andrew', 'james'])

    def test_query_metadata_returns_correct_null_values(self):
        query = '''
        select 'andrew' as name, null as age
        union all
        select null as name, null as age
        '''

        bqr = BigQueryResult(query)
        result_metadata = bqr.metadata()
        null_names = result_metadata['null_values']['name']
        self.assertEquals(null_names, 1)

        null_age = result_metadata['null_values']['age']
        self.assertEquals(null_age, 2)

    def test_query_metadata_returns_correct_value_counts(self):
        query = '''
        select 'andrew' as name, 20 as age
        union all
        select 'john' as name, 20 as age
        '''

        bqr = BigQueryResult(query)
        result_metadata = bqr.metadata()
        value_counts = result_metadata['value_counts']
        name_value_counts = value_counts['name']
        self.assertEquals(name_value_counts, {'john': 1, 'andrew': 1})

        age_value_counts = value_counts['age']
        self.assertEquals(age_value_counts, {20: 2})



import argparse
import logging


def main():
    from bigquery_validator import BigQueryValidator, print_failure, print_success

    parser = argparse.ArgumentParser()
    parser.add_argument('params', nargs='*')

    args = parser.parse_args()
    function = args.params[0]
    param = args.params[1]

    bigquery_validator = BigQueryValidator()
    if function == 'render_templated_query':
        result = bigquery_validator.render_templated_query(param)
        print(result)
    # elif function == 'dry_run_query':
    #     valid_query = bigquery_validator.dry_run_query(param)
    #     result = f'Query is {valid_query}'
    elif function == 'validate_query':
        valid_query, message = bigquery_validator.validate_query(param)
        if valid_query:
            print_success(f'Valid query. {message}', end='\n')
        else:
            print_failure(f'Invalid query. {message}', end='\n')
    elif function == 'validate_query_from_file':
        valid_query = bigquery_validator.validate_query_from_file(param)
        if valid_query:
            print_success('Valid query')
        else:
            print_failure('Invalid query')
    elif function == 'auto_validate_query_from_file':
        bigquery_validator.auto_validate_query_from_file(param)
    else:
        raise ValueError('Invalid argument passed for function')
    # TODO: Accept arg for print to console or print to file


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.WARNING)
    main()



from datetime import datetime, timedelta

today = datetime.now()
yesterday = datetime.now() - timedelta(days=1)


def datetime_to_ds(dt):
    return dt.strftime('%Y-%m-%d')


def datetime_to_ds_nodash(dt):
    return dt.strftime('%Y%m%d')


def get_default_params():
    return {
        'ds': datetime_to_ds(yesterday),
        'ds_nodash': datetime_to_ds_nodash(yesterday),
        'tomorrow_ds_nodash': datetime_to_ds_nodash(today),
        'tomorrow_ds': datetime_to_ds(today)
    }



from google.cloud import bigquery
import pandas as pd

from bigquery_validator import BigQueryValidator


class BigQueryResult:
    """Execute a BigQuery query and extract metadata such total rows and column, unique values in each column, null value
    counts for each column and count of each value in a column.
    """

    def __init__(self,
                 query=None,
                 file_path=None,
                 auto_execute=True,
                 use_query_cache=False):
        self.bq_client = bigquery.Client()
        self.bigquery_validator = BigQueryValidator()

        if file_path is not None and query is not None:
            raise ValueError(
                'Both a query and path to a .sql file were specified.'
                ' Please specify only one of these.')
        elif file_path is None and query is None:
            raise ValueError('A query or path to a .sql file must be specified')
        elif query is not None:
            self.query = self.bigquery_validator.render_templated_query(query)
        else:
            self.query = self.bigquery_validator.render_templated_query_from_file(file_path)

        self.auto_execute = auto_execute
        self.result = []
        self.use_query_cache = use_query_cache

        if self.auto_execute:
            self.execute()

    def execute(self):
        """Execute BigQuery query and returns result as a list of Python dicts"""
        try:
            job_config = bigquery.QueryJobConfig(use_query_cache=self.use_query_cache)
            query_result = self.bq_client.query(self.query, job_config=job_config)
            # todo store query cost
            self.result = [dict(r) for r in query_result]
            return self.result
        except Exception as e:
            print(e)

    def result(self):
        """Return the result of the query as a Python list"""
        return self.result

    def dataframe(self):
        """Return the result of the query as a dataframe"""
        return pd.DataFrame(self.result)

    def metadata(self):
        """Return additional data related to the result of a query including:
        - nrows: total number of rows in the query result
        - ncols: total number of columns in the query result
        - null_values: count of null values for each column
        - unique_values: the unique values for each column in the result
        - value_counts:  counts of unique values for each column
        """
        df = self.dataframe()

        metadata = {
            'columns': list(df.columns.values),
            'nrows': df.shape[0],
            'ncols': df.shape[1],
            'null_values': {},
            'unique_values': {},
            'value_counts': {}
        }

        for column in df.columns:
            unique_values = list(df[column].unique())
            metadata['unique_values'][column] = unique_values

            null_value_counts = df[column].isna().sum()
            metadata['null_values'][column] = null_value_counts

            value_counts_df = df[column].value_counts()
            value_counts = value_counts_df.to_dict()
            metadata['value_counts'][column] = value_counts
        return metadata



import unittest

from bigquery_validator.bigquery_validator import BigQueryValidator


class BigqueryValidatorTest(unittest.TestCase):

    bigquery_validator = BigQueryValidator()

    def test_valid_query_returns_true(self):
        query = "SELECT count(*) FROM `{{ params.project }}.samples.github_timeline`"
        valid_sql, _ = self.bigquery_validator.validate_query(query)
        self.assertTrue(valid_sql)

    def test_bad_query_returns_false(self):
        query = "SELECT count(*) ROM `bigquery-public-data.samples.github_timeline`"
        bad_sql, _ = self.bigquery_validator.validate_query(query)
        self.assertFalse(bad_sql)

    def test_valid_query_from_file_returns_true(self):
        valid_sql, _ = self.bigquery_validator.validate_query_from_file("./valid_query.sql")
        self.assertTrue(valid_sql)

    def test_bad_query_from_file_returns_false(self):
        bad_sql, _ = self.bigquery_validator.validate_query_from_file("./bad_query.sql")
        self.assertFalse(bad_sql, 'assert_bad_sql_from_file_fails_validation')

    def test_parameterised_query_is_formatted_correctly(self):
        templated_query = "select count(*) from `{{ params.project }}.samples.github_timeline`"
        rendered_query = self.bigquery_validator.render_templated_query(templated_query)
        self.assertEquals(rendered_query, 'select count(*) from `bigquery-public-data.samples.github_timeline`')

    def test_parameterised_query_from_file_is_formatted_correctly(self):
        rendered_query = self.bigquery_validator.render_templated_query_from_file("./test_param_query.sql")
        self.assertEquals(rendered_query, 'select count(*) from `bigquery-public-data.samples.github_timeline`')

    def test_extra_params_are_loaded_from_file(self):
        test_param = self.bigquery_validator.params.get('environment')
        self.assertEqual(test_param, 'test', 'assert_extra_param_exists')

    def test_message_returns_correct_expected_processing_size(self):
        expected_message = "This query will process 395.51 MB."
        query = "SELECT repository_url, repository_has_downloads, repository_created_at, repository_has_issues, " \
                "repository_forks FROM `bigquery-public-data.samples.github_timeline`"
        _, message = self.bigquery_validator.validate_query(query)
        self.assertEquals(message, expected_message, 'assert_message_returns_correct_expected_processing_size')

    def test_query_costs_less_than_1_gb(self):
        bigquery_validator = BigQueryValidator(return_query_cost_as_dict=True)
        query = "SELECT repository_url, repository_has_downloads, repository_created_at, repository_has_issues, " \
                "repository_forks FROM `bigquery-public-data.samples.github_timeline`"
        _, query_cost = bigquery_validator.validate_query(query)
        query_cost_gb = query_cost['gb']
        query_cost_mb = query_cost['mb']
        self.assertLess(query_cost_gb, 1, 'assert_query_costs_less_than_1_gigabyte')
        self.assertGreater(query_cost_mb, 100, 'assert_query_costs_greater_than_100_megabyte')
